using System;
using System.Concurrency;
using Microsoft.Maestro;

public domain TaxService
{
    public channel TaxCalculator
    {
        input int Income;
        output int Tax;
    }

    public abstract agent TaxAgent : channel TaxCalculator
    {
        protected abstract double GetTaxRate();

        public TaxAgent()
        {
            var income = receive(this.PrimaryChannel::Income);
            var tax = (int)(income * GetTaxRate());
            this.PrimaryChannel::Tax <-- tax;
        }
    }        
    
    public agent TaxHeaven : TaxAgent
    {
        protected override double GetTaxRate() { return 0.1; }
    }
    
    public agent TaxHell : TaxAgent
    {
        protected override double GetTaxRate() { return 0.9; }
    }

    public TaxService()
    {
        host TaxHeaven "heaven";
        host TaxHell "hell";
    }    
}

public domain Program
{
    TaxService taxService = new TaxService();
    
    agent MainAgent : channel Microsoft.Maestro.Application
    {
        public MainAgent()
        {
            Console.Write("Input your income:$");
            var income = int.Parse(Console.ReadLine());
            
            var taxCalculatorHeaven = new TaxService.TaxCalculator("heaven");
            var taxCalculatorHell = new TaxService.TaxCalculator("hell");
            
            taxCalculatorHeaven::Income <-- income;
            taxCalculatorHell::Income <-- income;
            
            // Do something else while the taxes are being calculated...
            
            var taxInHeaven = receive(taxCalculatorHeaven::Tax);
            var taxInHell = receive(taxCalculatorHell::Tax);

            Console.WriteLine("Your tax in heaven would be ${0}, your tax in hell would be ${1}", taxInHeaven, taxInHell);
            
            Console.ReadLine();
            PrimaryChannel::Done <-- Signal.Value;
        }
    }
}

