using System;
using System.Concurrency;
using Microsoft.Maestro;

public domain TaxService
{
    public TaxService()
    {
        host TaxHeaven "heaven";
        host TaxHell "hell";
    }
    
    public channel TaxCalculator
    {
        input int Income : int; // request-reply port: takes int, returns int
    }
    
    public abstract agent TaxAgent : channel TaxCalculator
    {
        protected abstract double GetTaxRate();
        
        public TaxAgent()
        {
            this.PrimaryChannel::Income ==> CalculateTax; // forwarding expression
        }
        
        private int CalculateTax(int income)
        {
            return (int)(income * GetTaxRate());
        }
    }        
    
    public agent TaxHeaven : TaxAgent
    {
        protected override double GetTaxRate() { return 0.1; }
    }
    
    public agent TaxHell : TaxAgent
    {
        protected override double GetTaxRate() { return 0.9; }
    }    
   
}

public domain Program
{
    TaxService taxService = new TaxService();
    
    agent MainAgent : channel Microsoft.Maestro.Application
    {
        public MainAgent()
        {
            Console.WriteLine("John and Mary filing separately");
            Console.Write("Input John's income:$");
            var income1 = int.Parse(Console.ReadLine());
            
            Console.Write("Input Mary's income:$");
            var income2 = int.Parse(Console.ReadLine());            
            
            var taxCalculatorHeaven = new TaxService.TaxCalculator("heaven");
            var taxCalculatorHell = new TaxService.TaxCalculator("hell");
            
            var correlatorHeaven1 = taxCalculatorHeaven::Income <-- income1;
            var correlatorHell1 = taxCalculatorHell::Income <-- income1;
            
            var correlatorHeaven2 = taxCalculatorHeaven::Income <-- income2;
            var correlatorHell2 = taxCalculatorHell::Income <-- income2;            
            
            // Do MORE useful work while the taxes are being calculated...
            
            var taxInHeaven1 = receive(correlatorHeaven1);
            var taxInHell1 = receive(correlatorHell1);
            
            var taxInHeaven2 = receive(correlatorHeaven2);
            var taxInHell2 = receive(correlatorHell2);            

            Console.WriteLine("John's tax in heaven would be ${0}, John's tax in hell would be ${1}", taxInHeaven1, taxInHell1);
            Console.WriteLine("Mary's tax in heaven would be ${0}, Mary's tax in hell would be ${1}", taxInHeaven2, taxInHell2);
            
            Console.ReadLine();
            PrimaryChannel::Done <-- Signal.Value;
        }
    }
}

